// -*- coding: utf-8 -*-
// vim:fileencoding=utf-8
/**
@author Devyn Collier Johnson <DevynCJohnson@Gmail.com>
@copyright LGPLv3

@brief DCJ-Encryption Algorithm (16384-byte/131072-bit encryption)
@version 2016.02.09

@section LICENSE
GNU Lesser General Public License v3
Copyright (c) Devyn Collier Johnson, All rights reserved.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3.0 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library.

@section INFO


* COMPILE *

To build the program, execute the below code for the target system

# Clang; General Build
LD=llvm-link; AS=llvm-as; RANLIB=llvm-ranlib; AR=llvm-ar; clang -Qunused-arguments -Wall -Wextra -pedantic -std=c11 -O3 -Wl,-O3 -g0 -ggdb0 -Wl,-s -s -funroll-loops -fwrapv -fno-stack-protector -fstrict-aliasing -fstrict-overflow -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -o ./dcj ./dcj_encryption.c

# Clang; General Build (POSIX and/or Unix)
LD=llvm-link; AS=llvm-as; RANLIB=llvm-ranlib; AR=llvm-ar; clang -Qunused-arguments -Wall -Wextra -pedantic -std=c11 -O3 -Wl,-O3 -g0 -ggdb0 -Wl,-s -s -funroll-loops -ftree-vectorize -ffunction-sections -fdata-sections -fwrapv -fstack-protector-strong -fstrict-aliasing -fstrict-overflow -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -o ./dcj ./dcj_encryption.c && strip --strip-all --strip-debug --strip-unneeded --discard-locals --remove-section=.bss --remove-section=.eh_frame --remove-section=.eh_frame_hdr --remove-section=.note --remove-section=.note.ABI-tag --remove-section=.note.gnu.build-id --remove-section=.comment --remove-section=.gnu.version --remove-section=.jcr --remove-section=.line --remove-section=.stab --remove-section=.stabstr ./dcj

# GNU-GCC; General Build (POSIX and/or Unix)
gcc -Wall -Wextra -pedantic -std=c11 -O3 -Wl,-O3 -g0 -ggdb0 -Wl,-s -s -fwhole-program -funroll-loops -fwrapv -fstack-protector-strong -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -fno-signaling-nans -o ./dcj ./dcj_encryption.c

# GNU-GCC; 32-bit Intel Build (POSIX and/or Unix)
gcc -flto -fuse-linker-plugin -Wall -Wextra -pedantic -std=c11 -mtune=intel -m32 -minline-all-stringops -momit-leaf-frame-pointer -maccumulate-outgoing-args -O3 -Wl,-O3 -g0 -ggdb0 -Wl,-s -s -funroll-loops -funsafe-loop-optimizations -ftree-vectorize -ffunction-sections -fdata-sections -fwrapv -fstack-protector-strong -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -fno-signaling-nans -o ./dcj ./dcj_encryption.c && strip --strip-all --strip-debug --strip-unneeded --discard-locals --remove-section=.bss --remove-section=.eh_frame --remove-section=.eh_frame_hdr --remove-section=.note --remove-section=.note.ABI-tag --remove-section=.note.gnu.build-id --remove-section=.comment --remove-section=.gnu.version --remove-section=.jcr --remove-section=.line --remove-section=.stab --remove-section=.stabstr ./dcj

# GNU-GCC; 64-bit Intel Build (POSIX and/or Unix)
gcc -flto -fuse-linker-plugin -Wall -Wextra -pedantic -std=c11 -mtune=intel -mmmx -msse2 -m64 -minline-all-stringops -momit-leaf-frame-pointer -maccumulate-outgoing-args -O3 -Wl,-O3 -g0 -ggdb0 -Wl,-s -s -funroll-loops -funsafe-loop-optimizations -ftree-vectorize -ffunction-sections -fdata-sections -fwrapv -fstack-protector-strong -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -fno-signaling-nans -o ./dcj ./dcj_encryption.c && strip --strip-all --strip-debug --strip-unneeded --discard-locals --remove-section=.bss --remove-section=.eh_frame --remove-section=.eh_frame_hdr --remove-section=.note --remove-section=.note.ABI-tag --remove-section=.note.gnu.build-id --remove-section=.comment --remove-section=.gnu.version --remove-section=.jcr --remove-section=.line --remove-section=.stab --remove-section=.stabstr ./dcj

# GNU-GCC; Generic x86-32 Build (POSIX and/or Unix)
gcc -flto -fuse-linker-plugin -Wall -Wextra -pedantic -std=c11 -mtune=generic -m32 -minline-all-stringops -momit-leaf-frame-pointer -maccumulate-outgoing-args -O3 -Wl,-O3 -g0 -ggdb0 -Wl,-s -s -funroll-loops -funsafe-loop-optimizations -ftree-vectorize -ffunction-sections -fdata-sections -fwrapv -fstack-protector-strong -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -fno-signaling-nans -o ./dcj ./dcj_encryption.c && strip --strip-all --strip-debug --strip-unneeded --discard-locals --remove-section=.bss --remove-section=.eh_frame --remove-section=.eh_frame_hdr --remove-section=.note --remove-section=.note.ABI-tag --remove-section=.note.gnu.build-id --remove-section=.comment --remove-section=.gnu.version --remove-section=.jcr --remove-section=.line --remove-section=.stab --remove-section=.stabstr ./dcj

# GNU-GCC; Generic x86-64 Build (POSIX and/or Unix)
gcc -flto -fuse-linker-plugin -Wall -Wextra -pedantic -std=c11 -mtune=generic -m64 -minline-all-stringops -momit-leaf-frame-pointer -maccumulate-outgoing-args -O3 -Wl,-O3 -g0 -ggdb0 -Wl,-s -s -funroll-loops -funsafe-loop-optimizations -ftree-vectorize -ffunction-sections -fdata-sections -fwrapv -fstack-protector-strong -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -fno-signaling-nans -o ./dcj ./dcj_encryption.c && strip --strip-all --strip-debug --strip-unneeded --discard-locals --remove-section=.bss --remove-section=.eh_frame --remove-section=.eh_frame_hdr --remove-section=.note --remove-section=.note.ABI-tag --remove-section=.note.gnu.build-id --remove-section=.comment --remove-section=.gnu.version --remove-section=.jcr --remove-section=.line --remove-section=.stab --remove-section=.stabstr ./dcj

# GNU-GCC; Generic x86-64 Musl-libc Build (POSIX and/or Unix)
CC="musl-gcc"; musl-gcc -D__MUSL__ -static -fwhole-program -Wall -Wextra -pedantic -Wformat -Werror=format-security -Wformat-nonliteral -Wformat-signedness -Wformat-y2k -Wmissing-declarations -Wmissing-prototypes -Wstrict-prototypes -Wstrict-overflow=5 -Wuninitialized -Wshadow -Wpointer-arith -Wcast-align -Wcast-qual -Wwrite-strings -Wredundant-decls -Wnested-externs -Winline -Wconversion -Wbad-function-cast -Wdouble-promotion -Wmissing-include-dirs -Wswitch-default -Wswitch-enum -Wswitch-bool -Wsync-nand -Winit-self -Wsuggest-final-types -Wsuggest-final-methods -Wundef -Wdate-time -Wjump-misses-init -Wlogical-op -Wopenmp-simd -Wpacked -Winvalid-pch -Wunsafe-loop-optimizations -Wvector-operation-performance -Wdisabled-optimization -Wstack-protector -std=c11 -mtune=generic -m64 -minline-all-stringops -momit-leaf-frame-pointer -maccumulate-outgoing-args -O3 -Wl,-O3 -g0 -ggdb0 -Wl,-s -s -funroll-loops -funsafe-loop-optimizations -ffunction-sections -fdata-sections -fwrapv -fstack-protector-strong -fomit-frame-pointer -fno-signed-zeros -fno-math-errno -fno-signaling-nans -o ./dcj_musl ./dcj_encryption.c && strip --strip-all --strip-debug --strip-unneeded --discard-locals --remove-section=.bss --remove-section=.eh_frame --remove-section=.eh_frame_hdr --remove-section=.note --remove-section=.note.ABI-tag --remove-section=.note.gnu.build-id --remove-section=.comment --remove-section=.gnu.version --remove-section=.jcr --remove-section=.line --remove-section=.stab --remove-section=.stabstr ./dcj_musl

# MingW; Generic x86-32 Build (Windows)
CC="i686-w64-mingw32-gcc -I/usr/i686-w64-mingw32/include -Wl,-rpath=-L/usr/i686-w64-mingw32/lib"; CBUILD="i686-w64-mingw32-gcc"; CHOST="i686-w64-mingw32-gcc"; STRIP="i686-w64-mingw32-strip"; LD="i686-w64-mingw32-ld -L/usr/i686-w64-mingw32/lib"; AS="i686-w64-mingw32-as"; RANLIB="i686-w64-mingw32-ranlib"; AR="i686-w64-mingw32-ar"; NM="i686-w64-mingw32-nm"; DLLTOOL="i686-w64-mingw32-dlltool"; WINDRES="i686-w64-mingw32-windres"; INCLUDES="-I/usr/i686-w64-mingw32/include"; LIBS="-L/usr/i686-w64-mingw32/lib"; i686-w64-mingw32-gcc -I/usr/i686-w64-mingw32/include -Wl,-rpath=-L/usr/i686-w64-mingw32/lib -Wall -Wextra -pedantic -fms-extensions -mconsole -mwin32 -std=c11 -mtune=generic -m32 -O3 -Wl,-O3 -g0 -ggdb0 -Wl,-s -s -fwhole-program -funroll-loops -fwrapv -fno-stack-protector -D_FORTIFY_SOURCE=0 -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -fno-signaling-nans -o ./dcj.exe ./dcj_encryption.c && i686-w64-mingw32-strip --strip-debug --strip-unneeded ./dcj.exe

# MingW; Generic x86-64 Build (Windows)
CC="x86_64-w64-mingw32-gcc -I/usr/x86_64-w64-mingw32/include -Wl,-rpath=-L/usr/x86_64-w64-mingw32/lib"; CBUILD="x86_64-w64-mingw32-gcc"; CHOST="x86_64-w64-mingw32-gcc"; STRIP="x86_64-w64-mingw32-strip"; LD="x86_64-w64-mingw32-ld -L/usr/x86_64-w64-mingw32/lib"; AS="x86_64-w64-mingw32-as"; RANLIB="x86_64-w64-mingw32-ranlib"; AR="x86_64-w64-mingw32-ar"; NM="x86_64-w64-mingw32-nm"; DLLTOOL="x86_64-w64-mingw32-dlltool"; WINDRES="x86_64-w64-mingw32-windres"; INCLUDES="-I/usr/x86_64-w64-mingw32/include"; LIBS="-L/usr/x86_64-w64-mingw32/lib"; x86_64-w64-mingw32-gcc -I/usr/x86_64-w64-mingw32/include -Wl,-rpath=-L/usr/x86_64-w64-mingw32/lib -Wall -Wextra -pedantic -fms-extensions -mconsole -mwin32 -std=c11 -mtune=generic -m64 -O3 -Wl,-O3 -g0 -ggdb0 -Wl,-s -s -fwhole-program -funroll-loops -fwrapv -fno-stack-protector -D_FORTIFY_SOURCE=0 -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -fno-signaling-nans -o ./dcj.exe ./dcj_encryption.c && x86_64-w64-mingw32-strip --strip-debug --strip-unneeded ./dcj.exe

# MingW; 32-bit Intel Build (Windows)
CC="i686-w64-mingw32-gcc -I/usr/i686-w64-mingw32/include -Wl,-rpath=-L/usr/i686-w64-mingw32/lib"; CBUILD="i686-w64-mingw32-gcc"; CHOST="i686-w64-mingw32-gcc"; STRIP="i686-w64-mingw32-strip"; LD="i686-w64-mingw32-ld -L/usr/i686-w64-mingw32/lib"; AS="i686-w64-mingw32-as"; RANLIB="i686-w64-mingw32-ranlib"; AR="i686-w64-mingw32-ar"; NM="i686-w64-mingw32-nm"; DLLTOOL="i686-w64-mingw32-dlltool"; WINDRES="i686-w64-mingw32-windres"; INCLUDES="-I/usr/i686-w64-mingw32/include"; LIBS="-L/usr/i686-w64-mingw32/lib"; i686-w64-mingw32-gcc -I/usr/i686-w64-mingw32/include -Wl,-rpath=-L/usr/i686-w64-mingw32/lib -Wall -Wextra -pedantic -fms-extensions -mconsole -mwin32 -std=c11 -mtune=intel -m32 -O3 -Wl,-O3 -g0 -ggdb0 -Wl,-s -s -fwhole-program -funroll-loops -fwrapv -fno-stack-protector -D_FORTIFY_SOURCE=0 -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -fno-signaling-nans -o ./dcj.exe ./dcj_encryption.c && i686-w64-mingw32-strip --strip-debug --strip-unneeded ./dcj.exe

# MingW; 64-bit Intel Build (Windows)
CC="x86_64-w64-mingw32-gcc -I/usr/x86_64-w64-mingw32/include -Wl,-rpath=-L/usr/x86_64-w64-mingw32/lib"; CBUILD="x86_64-w64-mingw32-gcc"; CHOST="x86_64-w64-mingw32-gcc"; STRIP="x86_64-w64-mingw32-strip"; LD="x86_64-w64-mingw32-ld -L/usr/x86_64-w64-mingw32/lib"; AS="x86_64-w64-mingw32-as"; RANLIB="x86_64-w64-mingw32-ranlib"; AR="x86_64-w64-mingw32-ar"; NM="x86_64-w64-mingw32-nm"; DLLTOOL="x86_64-w64-mingw32-dlltool"; WINDRES="x86_64-w64-mingw32-windres"; INCLUDES="-I/usr/x86_64-w64-mingw32/include"; LIBS="-L/usr/x86_64-w64-mingw32/lib"; x86_64-w64-mingw32-gcc -I/usr/x86_64-w64-mingw32/include -Wl,-rpath=-L/usr/x86_64-w64-mingw32/lib -Wall -Wextra -pedantic -fms-extensions -mconsole -mwin32 -std=c11 -mtune=intel -mmmx -msse2 -m64 -O3 -Wl,-O3 -g0 -ggdb0 -Wl,-s -s -fwhole-program -funroll-loops -fwrapv -fno-stack-protector -D_FORTIFY_SOURCE=0 -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -fno-signaling-nans -o ./dcj.exe ./dcj_encryption.c && x86_64-w64-mingw32-strip --strip-debug --strip-unneeded ./dcj.exe

# GNU-GCC; 64-bit Haswell Build (Linux)
gcc -flto -fuse-linker-plugin -Wall -Wextra -pedantic -std=c11 -march=haswell -m64 -mcrc32 -msahf -mcx16 -mvzeroupper -mcld -minline-all-stringops -momit-leaf-frame-pointer -maccumulate-outgoing-args -O3 -Wl,-O3 -Wl,-z,relro -Wl,-s -s -g0 -ggdb0 -funroll-loops -funsafe-loop-optimizations -ftree-vectorize -ffunction-sections -fdata-sections -fwrapv -fstack-protector-strong -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -fno-signaling-nans -o ./dcj ./dcj_encryption.c && strip --strip-all --strip-debug --strip-unneeded --discard-locals --remove-section=.bss --remove-section=.eh_frame --remove-section=.eh_frame_hdr --remove-section=.note --remove-section=.note.ABI-tag --remove-section=.note.gnu.build-id --remove-section=.comment --remove-section=.gnu.version --remove-section=.jcr --remove-section=.line --remove-section=.stab --remove-section=.stabstr ./dcj

# GNU-GCC; Faster 64-bit Quad-core Haswell Build (Linux)
gcc -ftree-parallelize-loops=4 -floop-parallelize-all -ffast-math -funsafe-math-optimizations -flto -fuse-linker-plugin -Wall -Wextra -pedantic -Wformat -Werror=format-security -Wformat-nonliteral -Wformat-signedness -Wformat-y2k -Wmissing-declarations -Wmissing-prototypes -Wstrict-prototypes -Wstrict-overflow=5 -Wuninitialized -Wshadow -Wpointer-arith -Wcast-align -Wcast-qual -Wwrite-strings -Wredundant-decls -Wnested-externs -Winline -Wconversion -Wbad-function-cast -Wdouble-promotion -Wmissing-include-dirs -Wswitch-default -Wswitch-enum -Wswitch-bool -Wsync-nand -Winit-self -Wsuggest-final-types -Wsuggest-final-methods -Wundef -Wdate-time -Wjump-misses-init -Wlogical-op -Wopenmp-simd -Wpacked -Winvalid-pch -Wunsafe-loop-optimizations -Wvector-operation-performance -Wdisabled-optimization -Wstack-protector -std=c11 -march=haswell -m64 -mcrc32 -msahf -mcx16 -mvzeroupper -mcld -minline-all-stringops -momit-leaf-frame-pointer -maccumulate-outgoing-args -O3 -Wl,-O3 -Wl,-z,relro -Wl,-s -s -g0 -ggdb0 -funroll-loops -funsafe-loop-optimizations -ftree-vectorize -ffunction-sections -fdata-sections -fwrapv -fstack-protector-strong -fomit-frame-pointer -ffinite-math-only -fno-signed-zeros -fno-math-errno -fno-signaling-nans -o ./dcj ./dcj_encryption.c && strip --strip-all --strip-debug --strip-unneeded --discard-locals --remove-section=.bss --remove-section=.eh_frame --remove-section=.eh_frame_hdr --remove-section=.note --remove-section=.note.ABI-tag --remove-section=.note.gnu.build-id --remove-section=.comment --remove-section=.gnu.version --remove-section=.jcr --remove-section=.line --remove-section=.stab --remove-section=.stabstr ./dcj


* USAGE *

A new key is generated for each encrypted file; existing files with the same name will be over-written
To encrypt a file - ./dcj -e ./key_file ./file ./encrypted_file
To decrypt a file - ./dcj -d ./key_file ./encrypted_file ./decrypted_file

Parameters
-e|-d KEY_FILE INPUT_FILE OUTPUT_FILE

-e: Encrypt
-d: Decrypt

*/


/* MACROS */


#if (defined(__GNUC__) && !defined(__clang__))
#   pragma GCC optimize (3)
#   pragma GCC optimize ("unroll-loops")
#   pragma GCC optimize ("wrapv")
#   define __IVDEP__  // ivdep: Vectorize loops
#elif defined(__INTEL_COMPILER)
#   pragma intel optimization_level 3
#endif


// C STANDARD
#ifndef _ISOC11_SOURCE  // C11
#   define _ISOC11_SOURCE
#endif
#ifndef __STDC_VERSION__
#   define __STDC_VERSION__   201112L  // C11
#endif
// ARCHITECTURE
#if (defined(__thumb__) && defined(__ARM_EABI__)) || defined(__arm__)
#   define ARCHARM
#endif
// OPERATING-SYSTEM MACROS
#ifdef EPLAN9  // Plan 9
#   define OSPLAN9
#elif defined(_AIX)
#   define OSAIX
#   define OSPOSIX
#   define OSUNIX
#elif (defined(__CYGWIN__) || defined(__CYGWIN32__))  // Cygwin
#   define OSCYGWIN
#   define OSPOSIX
#elif defined(__gnu_hurd__)  // GNU-Hurd
#   define OSHURD
#   define OSPOSIX
#elif (defined(_hpux) || defined(hpux) || defined(__hpux))  // HP-Unix
#   define OSHPUX
#   define OSPOSIX
#   define OSUNIX
#elif defined(__minix)  // Minix
#   define OSMINIX
#   define OSPOSIX
#elif (defined(__ANDROID__) && !defined(__gnu_linux__))  // Android
#   define OSANDROID
#   define OSLINUX
#   define OSPOSIX
#   ifdef _GNU_SOURCE  // Includes: ISO C89, ISO C99, POSIX.1, POSIX.2, BSD, SVID, X/Open, LFS, and GNU extensions
#      undef _GNU_SOURCE
#   endif
#   ifndef _FORTIFY_SOURCE
#      define _FORTIFY_SOURCE   (1)  // Adds lightweight checks that detect some buffer overflow errors
#   endif
#elif (defined(__linux__) || defined(linux) || defined(__linux)) || defined(__gnu_linux__) || defined(BOOST_OS_LINUX)  // Linux
#   define OSGNU
#   define OSLINUX
#   define OSPOSIX
#   ifndef _GNU_SOURCE
#      define _GNU_SOURCE  // Includes: ISO C89, ISO C99, POSIX.1, POSIX.2, BSD, SVID, X/Open, LFS, and GNU extensions
#   endif
#   ifndef _FORTIFY_SOURCE
#      define _FORTIFY_SOURCE   (1)  // Adds lightweight checks that detect some buffer overflow errors
#   endif
#elif (defined(__unix__) || defined(__unix) || defined(unix)) || (defined(__APPLE__) && defined(__MACH__))  // Unix-systems
#   define OSUNIX
#   define OSPOSIX
#   if (defined(macintosh) || defined(Macintosh) || defined(BOOST_OS_MACOS))
#      define OSMACOS
#   elif (defined(__APPLE__) && defined(__MACH__))
#      define OSOSX
#   elif (defined(__APPLE__) && !defined(__MACH__))
#      define OSAPPLE
#   elif (defined(__MACH__) && !defined(__APPLE__))
#      define OSNEXTSTEP
#   endif
#elif (defined(__MINGW32__) || defined(__MINGW64__) || defined(__WINNT__) || defined(__NT__) || defined(_WIN64) || defined(_WIN32) || defined(WIN32) || defined(__WIN32__) || defined(BOOST_OS_WINDOWS) || defined(_UWIN) || defined(__TOS_WIN__) || defined(__WINDOWS__) || defined(_WIN32_WCE) || defined(_MSC_VER))  // Windows
#   define OSWINDOWS
#   include <windows.h>
#   ifdef OSPOSIX
#      undef OSPOSIX  // Ensure `OSPOSIX` is not defined
#   endif
#   ifdef OSUNIX
#      undef OSUNIX  // Ensure `OSUNIX` is not defined
#   endif
#   ifdef _GNU_SOURCE
#      undef _GNU_SOURCE  // Ensure `_GNU_SOURCE` is not defined
#   endif
#   if (!defined(_FORTIFY_SOURCE) || _FORTIFY_SOURCE != 0)
#      define _FORTIFY_SOURCE   (0)  // Fix bug in MingW
#   endif
#endif
#ifndef _GNU_SOURCE
#   ifdef OSPOSIX  // Set special POSIX macros
#      define _POSIX_SOURCE   (200809L)  // Includes POSIX functionality
#      define _POSIX_C_SOURCE   (200809L)  // Includes POSIX functionality
#   endif
#   ifdef OSUNIX  // Set special UNIX macros
#      define _XOPEN_SOURCE   (700)  // Includes functionality described in the X/Open Portability Guide
#      define _XOPEN_VERSION   (700)  // Includes functionality described in the X/Open Portability Guide
#      define _XOPEN_SOURCE_EXTENDED  // Includes extra functions for the X/Open Unix brand
#   endif
#endif
#if (defined(__DragonFly__) || defined(__OpenBSD__) || defined(__NetBSD__) || defined(__FreeBSD__) || defined(_SYSTYPE_BSD) || defined(__bsdi__) || defined(BSD))  // BSD
#   undef _GNU_SOURCE
#   undef _POSIX_SOURCE
#   undef _POSIX_C_SOURCE
#   define _BSD_SOURCE
#   define _XOPEN_SOURCE   (700)  // Includes functionality described in the X/Open Portability Guide
#   define _XOPEN_VERSION   (700)  // Includes functionality described in the X/Open Portability Guide
#   define _XOPEN_SOURCE_EXTENDED  // Includes extra functions for the X/Open Unix brand
#   define OSBSD
#   define OSPOSIX
#   define OSUNIX
#elif (defined(__FreeBSD_kernel__ ) && defined(__GLIBC__ ))  // GNU-kFreeBSD
#   define OSGNU
#   define OSBSD
#   define OSPOSIX
#   define OSUNIX
#   define _GNU_SOURCE
#   define _POSIX_SOURCE   (200809L)  // Includes POSIX functionality
#   define _POSIX_C_SOURCE   (200809L)  // Includes POSIX functionality
#   ifdef _BSD_SOURCE
#      undef _BSD_SOURCE   // Deprecated in glibc
#   endif
#   if defined(__DragonFly__)
#      define OSDRAGONFLY
#   elif defined(__NetBSD__)
#      define OSNETBSD
#   elif defined(__OpenBSD__)
#      define OSOPENBSD
#   elif (defined(__FreeBSD__) || defined(__FreeBSD_kernel__))
#      define OSFREEBSD
#endif
#elif (defined(sun) || defined(__sun))  // Solaris, SunOS, etc.
#   define OSSUN
#   define OSPOSIX
#   define OSUNIX
#   if (defined(sun) || defined(__sun)) && (defined(__SVR4) || defined(__svr4__))  // Solaris
#      define OSSOLARIS
#   elif (defined(sun) || defined(__sun)) && !(defined(__SVR4) || defined(__svr4__))
#      define OSUNOS
#   endif
#endif
// SPECIAL INCLUDES
#ifdef OSUNIX
#   include <sys/param.h>
#endif
#if (defined(OSPOSIX) || defined(OSUNIX))
#   include <sys/types.h>  // http://pubs.opengroup.org/onlinepubs/9699919799/basedefs/sys_types.h.html
#   define SYS_TYPES_INCLUDED
#   include <unistd.h>
#endif


#include <errno.h>
#include <inttypes.h>  // Also includes <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#if ((defined(OSPOSIX) || defined(OSUNIX)) && !defined(__MUSL__))
#   include <sys/stat.h>  // http://pubs.opengroup.org/onlinepubs/9699919799/basedefs/sys_stat.h.html
#   include <sys/time.h>  // http://pubs.opengroup.org/onlinepubs/9699919799/basedefs/sys_time.h.html
#endif

#define DCJKEYLEN   ((uint64_t)16384)  // Key length (16384 bytes || 131072 bits)
#define DCJKEYLEN_M1   ((uint64_t)16383)  // Key length minus one
#define GRABDATACHAR   ((uint8_t)(((uint8_t)data[256] + 0x01) % (uint8_t)data[128]))
#define GRABKEYCHAR   ((uint8_t)(((uint8_t)key[512] + 0x01) % (uint8_t)key[128]))
#define MINDATASIZE   ((uint64_t)512)
#define uint   unsigned int
#define ulint   unsigned long int


/* PROTOTYPES */


size_t getfilesize_t(const char *filename);
#ifdef __GNUC__
void byte_rotl(uint8_t byte, const uint8_t num) __attribute__((flatten, hot));
void byte_rotr(uint8_t byte, const uint8_t num) __attribute__((flatten, hot));
void array_rotl(uint8_t *array, const uint64_t len, const uint64_t num_shifts) __attribute__((flatten, hot));
void array_rotr(uint8_t *array, const uint64_t len, const uint64_t num_shifts) __attribute__((flatten, hot));
void key_rotl(uint8_t *array, const uint8_t num_shifts) __attribute__((flatten, hot));
void key_rotr(uint8_t *array, const uint8_t num_shifts) __attribute__((flatten, hot));
#else
void byte_rotl(uint8_t byte, const uint8_t num);
void byte_rotr(uint8_t byte, const uint8_t num);
void array_rotl(uint8_t *array, const uint64_t len, const uint64_t num_shifts);
void array_rotr(uint8_t *array, const uint64_t len, const uint64_t num_shifts);
void key_rotl(uint8_t *array, const uint8_t num_shifts);
void key_rotr(uint8_t *array, const uint8_t num_shifts);
#endif
void writefile(const char *filename, const char *data, const size_t filesize);
int dcj_encrypt(uint8_t *data, const uint64_t datasize, char *pswdaddr);
int dcj_decrypt(uint8_t *data, const uint64_t datasize, uint8_t *key);
int main(int argc, char *argv[]);


/* FUNCTIONS */


#if ((defined(OSPOSIX) || defined(OSUNIX)) && !defined(__MUSL__))
size_t getfilesize_t(const char *filename) {
    // Return the size of a file (in size_t); POSIX systems only
    struct stat sb;
    if (stat(filename, &sb) != 0x00) {
        fprintf(stderr, "ERROR: `getfilesize()` failed to get the filesize of `%s`!\n%s\n", filename, strerror(errno));
        exit(EXIT_FAILURE);
    }
    return (size_t)sb.st_size;
}
#else
size_t getfilesize_t(const char *filename) {
    // Return the size of a file (in size_t)
    FILE *fileptr = fopen(filename, "r");
    if (fileptr == (FILE *)NULL) {
        fprintf(stderr, "ERROR: Failed to open file!\n%s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }
    if (fseek(fileptr, 0x00, SEEK_END) != 0x00) {
        fprintf(stderr, "ERROR: Failed to get filesize!\n%s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }
    size_t retval = (size_t)ftell(fileptr);
    fclose(fileptr);
    return retval;
}
#endif


#if ((defined(__x86_64__) || defined(__x86_64) || defined(__amd64__)) && !defined(__MUSL__))
void byte_rotl(uint8_t byte, const uint8_t num) {
    // Rotate the byte to the left `num` times
    __asm__ (  // AT&T Style Assembly
        "rol %1,%2;"
        : "=r" ( byte )  // %0: Output to C variable
        : "r" ( num ), "r" ( byte )  // %1 & %2: Input C variable
        :  // Clobbered Register
    );
}


void byte_rotr(uint8_t byte, const uint8_t num) {
    // Rotate the byte to the right `num` times
    __asm__ (  // AT&T Style Assembly
        "ror %1,%2;"
        : "=r" ( byte )  // %0: Output to C variable
        : "r" ( num ), "r" ( byte )  // %1 & %2: Input C variable
        :  // Clobbered Register
    );
}


#else


void byte_rotl(uint8_t byte, const uint8_t num) {
    // Rotate the byte to the left `num` times
    byte = (uint8_t)((uint8_t)(byte << (num & (uint8_t)0x07)) | (uint8_t)(byte >> ((-(num & (uint8_t)0x07)) & 0x07)));
}


void byte_rotr(uint8_t byte, const uint8_t num) {
    // Rotate the byte to the right `num` times
    byte = (uint8_t)((uint8_t)(byte >> (num & (uint8_t)0x07)) | (uint8_t)(byte << ((-(num & (uint8_t)0x07)) & 0x07)));
}
#endif


void array_rotl(uint8_t *array, const uint64_t len, const uint64_t num_shifts) {
    // Rotate an array to the left
    if ((len >= MINDATASIZE) && (num_shifts != 0x00)) {
        register uint64_t i, j;
        register uint8_t first;
        for (i = num_shifts; i != 0x00; --i) {
            first = array[0x00];
            for (j = 0x00; j < (len - 0x01); ++j) array[j] = array[j + 0x01];
            array[j] = first;
        }
    }
}


void array_rotr(uint8_t *array, const uint64_t len, const uint64_t num_shifts) {
    // Rotate an array to the right
    if ((len >= MINDATASIZE) && (num_shifts != 0x00)) {
        register uint64_t i, j;
        register uint8_t last;
        for (i = num_shifts; i != 0x00; --i) {
            last = array[len - 0x01];
            for (j = len; --j;) array[j] = array[j - 0x01];
            array[0x00] = last;
        }
    }
}


void key_rotl(uint8_t *array, const uint8_t num_shifts) {
    // Rotate key to the left
    if (num_shifts != 0x00) {
        register uint64_t j;
        register uint8_t i, first;
        for (i = num_shifts; i != 0x00; --i) {
            first = array[0x00];
            for (j = 0x00; j < DCJKEYLEN_M1; ++j) array[j] = array[j + 0x01];
            array[DCJKEYLEN_M1] = first;
        }
    }
}


void key_rotr(uint8_t *array, const uint8_t num_shifts) {
    // Rotate key to the right
    if (num_shifts != 0x00) {
        register uint64_t j;
        register uint8_t i, last;
        for (i = num_shifts; i != 0x00; --i) {
            last = array[DCJKEYLEN_M1];
            for (j = DCJKEYLEN; --j;) array[j] = array[j - 0x01];
            array[0x00] = last;
        }
    }
}


void writefile(const char *filename, const char *data, const size_t filesize) {
    // Write data to a file; buffered
    register FILE *wfileptr = fopen(filename, "wb");
    if (!wfileptr) {
        fprintf(stderr, "ERROR: Failed to open file!\n%s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }
    char *buffer = (char *)calloc((size_t)0x01, filesize);
    if (buffer == (char *)NULL) {
        fprintf(stderr, "ERROR: Failed to allocate memory to write file!\n");
        exit(EXIT_FAILURE);
    }
    setbuf(wfileptr, buffer);
    if ((fwrite(data, (size_t)0x01, filesize, wfileptr)) != filesize) {
        fclose(wfileptr);
        free(buffer);
        fprintf(stderr, "ERROR: Unable to write file!\n");
        exit(EXIT_FAILURE);
    }
    fclose(wfileptr);
    free(buffer);
}


int dcj_encrypt(uint8_t *data, const uint64_t datasize, char *pswdaddr) {
    // Encrypt a string using the DCJ Algorithm (16384 bytes || 131072 bits); returns `1` on success and `0` on failure
    if (datasize < MINDATASIZE) { return 0; }  // Ensure enough bytes are given
    else if ((pswdaddr == NULL) || (data == NULL)) { return 0; }  // Ensure NULL is not given
    /* SEED RANDOM-NUMBER GENERATOR */
#   if ((defined(OSPOSIX) || defined(OSUNIX)) && !defined(__MUSL__))
    struct timeval tmstrct;
    gettimeofday(&tmstrct, NULL);
    srand((uint)((uint)tmstrct.tv_usec | ((uint)tmstrct.tv_usec * (uint)tmstrct.tv_sec)));
#   else
    srand((uint)(((uint)time(NULL) + 0x01) % ((uint)time(NULL) * ((uint)clock() + 0x01))));
#   endif
    /* GENERATE KEY */
    const uint8_t charset[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789 \"\a\b\v\f\r\n\t<>,.?/~`!@#$%^&*()-_+=[]{}|\\':;\xCD\xCE\xCF\xD0\xD1";
    uint8_t *key = (uint8_t *)calloc((size_t)0x01, (size_t)DCJKEYLEN);
    if (key == NULL) {
        fprintf(stderr, "ERROR: Failed to allocate memory for key!\n");
        exit(EXIT_FAILURE);
    }
    register uint64_t i;
    for (i = (uint64_t)0x00; i < DCJKEYLEN; ++i) {  // Produce one char per loop
        key[i] = charset[(rand() % (int)107)];  // 107 is the length of `charset`
    }
    /* ENCRYPTION */
    array_rotr(data, datasize, (uint64_t)key[GRABKEYCHAR]);
    register uint64_t j = 0x00;
    for (i = datasize; --i;) {  // XOR Cipher
        if (j < DCJKEYLEN) { ++j; }
        else { j = 0x00; }
        data[i] ^= (uint8_t)key[j];
    }
    array_rotr(data, datasize, (uint64_t)key[GRABKEYCHAR]);
    j = 0x00;
    for (i = datasize; --i;) {  // XNOR Cipher
        if (j < DCJKEYLEN) { ++j; }
        else { j = 0x00; }
        data[i] = (uint8_t)(~(data[i] ^ key[j]));
    }
    key_rotl(key, (uint8_t)data[GRABDATACHAR]);
    j = 0x00;
    for (i = datasize; --i;) {  // XOR Cipher + Shift-Left
        if (j < DCJKEYLEN) { ++j; }
        else { j = 0x00; }
        data[i] ^= (uint8_t)key[j];
        byte_rotl(data[i], (uint8_t)key[(i + 0x01) % DCJKEYLEN]);
        data[i] ^= (uint8_t)key[j];
    }
    key_rotl(key, (uint8_t)data[GRABDATACHAR]);
    j = 0x00;
    for (i = datasize; --i;) {  // XOR Cipher
        if (j < DCJKEYLEN) { ++j; }
        else { j = 0x00; }
        data[i] ^= (uint8_t)key[j];
    }
    // END
    memcpy(pswdaddr, key, (size_t)DCJKEYLEN);  // Place key in desired location
    memset(key, 0x00, (size_t)DCJKEYLEN);  // Remove key from memory
    free(key);
    return 1;
}


int dcj_decrypt(uint8_t *data, const uint64_t datasize, uint8_t *key) {
    // Decrypt a string using the DCJ Algorithm (16384 bytes || 131072 bits); returns `1` on success and `0` on failure
    if (datasize < MINDATASIZE) { return 0; }  // Ensure enough bytes are given
    else if ((key == NULL) || (data == NULL)) { return 0; }  // Ensure NULL is not given
    register uint64_t i, j = 0x00;
    /* DECRYPTION */
    for (i = datasize; --i;) {  // XOR Cipher
        if (j < DCJKEYLEN) { ++j; }
        else { j = 0x00; }
        data[i] ^= key[j];
    }
    key_rotr((uint8_t *)key, (uint8_t)data[GRABDATACHAR]);
    j = 0x00;
    for (i = datasize; --i;) {  // XOR Cipher + Shift-Right
        if (j < DCJKEYLEN) { ++j; }
        else { j = 0x00; }
        data[i] ^= key[j];
        byte_rotr(data[i], key[(i + 0x01) % DCJKEYLEN]);
        data[i] ^= key[j];
    }
    key_rotr((uint8_t *)key, (uint8_t)data[GRABDATACHAR]);
    j = 0x00;
    for (i = datasize; --i;) {  // XNOR Cipher
        if (j < DCJKEYLEN) { ++j; }
        else { j = 0x00; }
        data[i] = (uint8_t)(((uint8_t)~data[i]) ^ key[j]);
    }
    array_rotl(data, datasize, (uint64_t)key[GRABKEYCHAR]);
    j = 0x00;
    for (i = datasize; --i;) {  // XOR Cipher
        if (j < DCJKEYLEN) { ++j; }
        else { j = 0x00; }
        data[i] ^= key[j];
    }
    array_rotl(data, datasize, (uint64_t)key[GRABKEYCHAR]);
    // END
    memset(key, 0x00, (size_t)DCJKEYLEN);  // Remove key from memory
    return 1;
}


int main(int argc, char *argv[]) {
    if (argc == 1) {  // Invalid arguments
        fprintf(stderr, "ERROR: Invalid number of arguments!\n");
        exit(EXIT_FAILURE);
    }
    else if (argc == 5 && ((strncmp(argv[1], "-e", (size_t)2) == 0) || (strncmp(argv[1], "-d", (size_t)2) == 0))) {
        /** ARGUMENTS
        argv[1] = (-e/-d) Encrypt/Decrypt
        argv[2] = key file
        argv[3] = path of file to Encrypt/Decrypt
        argv[4] = filepath for output data
        */
        const size_t filesize = getfilesize_t(argv[3]);
        FILE *fileptr = fopen(argv[3], "rb");
        if (!fileptr) {
            fprintf(stderr, "ERROR: Failed to open file!\n%s\n", strerror(errno));
            exit(EXIT_FAILURE);
        }
        uint8_t *file_contents = (uint8_t *)calloc((size_t)0x01, filesize);
        if (file_contents == (uint8_t *)NULL) {
            fprintf(stderr, "ERROR: Failed to allocate memory for input file!\n");
            exit(EXIT_FAILURE);
        }
        if ((size_t)(fread(file_contents, (size_t)0x01, filesize, fileptr)) != filesize) {
            free(file_contents);
            fprintf(stderr, "ERROR: Unable to read file!\n");
            exit(EXIT_FAILURE);
        }
        fclose(fileptr);
        if (strncmp(argv[1], "-e", (size_t)0x02) == 0x00) {  // Encrypt
            char *password = (char *)calloc((size_t)0x01, DCJKEYLEN);
            if (password == (char *)NULL) {
                fprintf(stderr, "ERROR: Failed to allocate memory for key!\n");
                exit(EXIT_FAILURE);
            }
            if (dcj_encrypt((uint8_t *)file_contents, (uint64_t)filesize, password) == 0) {
                free(password);
                free(file_contents);
                fprintf(stderr, "ERROR: Failed to encrypt data!\n");
                exit(EXIT_FAILURE);
            }
            writefile(argv[2], password, (size_t)DCJKEYLEN);  // Write key file
            memset(password, 0x00, DCJKEYLEN);  // Remove key from memory
            free(password);
        } else {  // Decrypt
            // Open key file
            FILE *key_fileptr = fopen(argv[2], "rb");
            if (!key_fileptr) {
                fprintf(stderr, "ERROR: Failed to open file!\n%s\n", strerror(errno));
                free(file_contents);
                exit(EXIT_FAILURE);
            }
            uint8_t *password = (uint8_t *)calloc((size_t)0x01, DCJKEYLEN);
            if (password == (uint8_t *)NULL) {
                fprintf(stderr, "ERROR: Failed to allocate memory for key!\n");
                exit(EXIT_FAILURE);
            }
            if ((fread(password, (size_t)0x01, (size_t)DCJKEYLEN, key_fileptr)) != (size_t)DCJKEYLEN) {
                free(password);
                free(file_contents);
                fprintf(stderr, "ERROR: Unable to read file!\n");
                exit(EXIT_FAILURE);
            }
            fclose(key_fileptr);
            // Decrypt
            if (dcj_decrypt((uint8_t *)file_contents, (uint64_t)filesize, (uint8_t *)password) == 0) {
                free(password);
                free(file_contents);
                fprintf(stderr, "ERROR: Failed to decrypt data!\n");
                exit(EXIT_FAILURE);
            }
            free(password);
        }
        writefile(argv[4], (char *)file_contents, (size_t)filesize);
        memset(file_contents, 0x00, filesize);  // Remove data from memory
        free(file_contents);
    } else  {  // Invalid arguments
        fprintf(stderr, "ERROR: Invalid number of arguments!\n");
        exit(EXIT_FAILURE);
    }
    /* END */
    exit(EXIT_SUCCESS);
}
